/*--------------------------------------------------------------------*/
/*                                                                    */
/*  Effective I/O Bandwidth (b_eff_io) Benchmark                      */
/*                                                                    */
/*  Author: Rolf Rabenseifner                                         */
/*                                                                    */
/*  Copyright (c) Rolf Rabenseifner, HLRS, University of Stuttgart    */
/*                                                                    */
/*  Further information: www.hlrs.de/mpi/b_eff_io/                    */
/*                                                                    */
#define Revision "1.1"
/*                                                                    */
#define Date     "June 29, 2000"
/*                                                                    */
/*--------------------------------------------------------------------*/

/* for ROMIO 1.0: #define WITHOUT_SHARED */

/* Compilation, execution and postprocessing -- examples:

   CRAY T3E:  Prerequisites: using moduls mpt
              Compilation on T3E :   
                cc -o b_eff_io -D WITHOUT_SHARED b_eff_io.c 
                cc -o b_eff_io -D WITHOUT_SHARED b_eff_io.c \
                   ../ufs_t3e/ad_ufs_open.o ../ufs_t3e/ad_ufs_read.o \
                   ../ufs_t3e/ad_ufs_write.o
                cc -o b_eff_io -D WITHOUT_SHARED b_eff_io.c ../ufs_t3e/ad_ufs_*.o 
              Execution: export MPI_BUFFER_MAX=4099 
              T3E-900 with 128 MB/processor and 512 PEs: 
                mpirun -np 64 ./b_eff_io -MB 128 -MT 65536 \
                       -p $SCRDIR -f b_eff_io_T3E900_064PE
              T3E-1200 with 512 MB/processor and 512: 
                mpirun -np <NP> ./b_eff_io -MB 512 -MT 262144 \
                       -p $SCRDIR -f b_eff_io_T3E1200_064PE

   SX-4:      Prerequisites: -
              Compilation on SX-4/32 with 256 MB/processor:
                mpicc -o b_eff_io b_eff_io.c -lm
              Execution: 
                mpirun -np 8 ./b_eff_io -MB 256 -MT 8192 \
                       -p $SCRDIR -f b_eff_io_SX4_08PE 

   Postprocessing (on local workstation):
                b_eff_io_eps 64 b_eff_io_T3E900_064PE 
                b_eff_io_eps 64 b_eff_io_T3E1200_064PE
                b_eff_io_eps  8 b_eff_io_SX4_08PE

   Outputfiles: b_eff_io_T3E900_064PE.sum       human readable summary
                                     .prot      full benchmark protocol
                                     _*_mono.eps   diagrams black/white
                                     _*_color.eps  colored, for slides
                                     _on1page.ps   summary sheet 
                Same for b_eff_io_T3E1200_064PE and b_eff_io_SX4_08PE.
*/ 

#define T_CRIT         900   /* see also HELP text */
#define T_PERCENTAGE  0.90   /* see also HELP text */

#define HELP \
\
" Synopsis: \n\n"\
\
"  mpicc -o b_eff_io <compile time options> b_eff_io.c -lm \n\n"\
\
"  mpirun -np <NP> ./b_eff_io <runtime options> \n\n"\
\
" General compile time options: \n\n"\
\
"  -D WITHOUT_SHARED            to substitute the shared file pointer \n"\
"                               by individual file pointers           \n"\
"                               (implies runtime option -noshared)    \n\n"\
\
" Runtime options: \n\n"\
\
"  -MB <number of megabytes memory per node>  (mandatory) \n"\
"                               A node is defined as the unit    \n"\
"                               used by or useable for one MPI   \n"\
"                               process. This value is used to   \n"\
"                               compute the maximum chunk size   \n"\
"                               for the patterns 1, 10, 18, 26 and 35.\n"\
"                               The maximum chunk size is defined     \n"\
"                               as max( 2MB, memory per node / 128).  \n\n"\
\
"  -MT <number of megabytes memory of the total system> (mandatory) \n"\
"                               This value is used to compute the   \n"\
"                               second criterion for valid results  \n"\
"                               for comparison of different systems:\n"\
"                               The data transferred with each      \n"\
"                               access method must be at least      \n"\
"                               90 percent of the total memory.   \n\n"\
\
"  -noshared                    to substitute the shared file pointer \n"\
"                               by individual file pointers in pattern\n"\
"                               type 1 (implied by the compile time   \n"\
"                               option -D WITHOUT_SHARED).            \n\n"\
\
"  -nounique                    to remove MPI_MODE_UNIQUE_OPEN   \n"\
"                               on each file opening             \n\n"\
\
"  -keep                        to keep all benchmarking files     \n"\
"                               on close after last pattern test   \n\n"\
\
"  -N <number of processes>     maximum number of MPI processes N  \n"\
"                               used for this benchmark            \n"\
"                               (default = size of MPI_COMM_WORLD) \n\n"\
\
"  -n <minimum no. of processes> The benchmark is done for        \n"\
"                               i = N, N/2, N/4, ... with i >= n  \n"\
"                               (default = N)                     \n\n"\
\
"  -T <scheduled time>          scheduled time for maximum number of  \n"\
"                               processes N (default = 900 [seconds]) \n\n"\
\
"  -t <minimum scheduled time>  scheduled time for any number of   \n"\
"                               processes i is  max(t, T * i / N)  \n"\
"                               (default = T)                      \n\n"\
\
"  -p <path of fast_filesystem> path of the filesystem that should be \n"\
"                               benchmarked, i.e.  where this bench-  \n"\
"                               marks should write its scratch files  \n"\
"                               (default is the current directory).   \n\n"\
\
"  -e <number of errors>        maximum of errors printed in each \n"\
"                               pattern (default = 1).            \n\n"\
\
"  -f <protocol files' prefix>  prefix of the protocol file and \n"\
"                               the summary file.               \n"\
"                               The protocol and summary will be named \n"\
"                                <prefix>.prot   and   <prefix>.sum    \n"\
"                               (default prefix = b_eff_io).           \n\n"\
\
" Remarks: \n"\
"   -- Already existing scratch files are automatically removed before\n"\
"      benchmarking is started. \n"\
"   -- If the result should be used for copmparing different systems, the\n"\
"      benchmark is only valid if the following criterions are reached:\n"\
"       (1)   T >= 900 sec, \n"\
"       (2)   T is chosen large enough, that the transferred amount of \n"\
"             data (by each access method) is at least 90 percent of   \n"\
"             the size of the total memory of the system, \n"\
"       (3)   the option -noshared is NOT used, and \n"\
"       (4)   no errors are reported. \n\n"\
\
" Postprocessing: \n\n"\
\
"   With b_eff_io_eps, one can select one section of summary protocol \n"\
"   file and generate diagrams. \n\n"\
\
"   Synopsis: b_eff_io_eps <number of processes> [ <protocol files' prefix> ] \n\n"\
\
"   Output: <protocol files' prefix>_write_mono.eps   (black/white, e.g.,\n"\
"           <protocol files' prefix>_rewrt_mono.eps    for publication)  \n"\
"           <protocol files' prefix>_read_mono.eps \n"\
"           <protocol files' prefix>_write_color.eps   (color, thick lines, \n"\
"           <protocol files' prefix>_rewrt_color.eps    e.g., for slides) \n"\
"           <protocol files' prefix>_read_color.eps \n\n"\
\
" See also:  www.hlrs.de/mpi/b_eff_io/ \n"


#include "mpi.h"

/* --- for printing */ 
#include <stdio.h>
/* --- for pow(), log(), exp(): */
#include <math.h>
/* --- for fast malloc(), time(): */
#include <sys/types.h>
/* --- for fast malloc(): */
#include <malloc.h>
/* --- for time(), ctime(): */
#include <time.h>
/* --- for drand48(), srand48(): */
#include <stdlib.h>
/* --- for uname(): */
#include <sys/utsname.h>
/* --- for strcmp(), strcat(): */
#include <string.h> 

#define ACCESSES   3           /* 0=write, 1=re-write, 2=read */
#define TYPE_0_PTRNS 9
#define TYPE_1_PTRNS 8
#define TYPE_2_PTRNS 8
#define TYPE_3_PTRNS 9
#define TYPE_4_PTRNS 9
#define PATTERNS  (TYPE_0_PTRNS+TYPE_1_PTRNS+TYPE_2_PTRNS+TYPE_3_PTRNS+TYPE_4_PTRNS)
#define TYPE_0_START 0
#define TYPE_1_START (TYPE_0_START+TYPE_0_PTRNS)
#define TYPE_2_START (TYPE_1_START+TYPE_1_PTRNS)
#define TYPE_3_START (TYPE_2_START+TYPE_2_PTRNS)
#define TYPE_4_START (TYPE_3_START+TYPE_3_PTRNS)
#define PAT_TYPES  5
#define POSITIONS  8 

#define max(a,b) ( (a)>(b) ?   (a)  : (b) )
#define min(a,b) ( (a)<(b) ?   (a)  : (b) )

#define M_PART -1
#define MB     *1024*1024 
#define kB     *1024
#define SAME   -1 

#define ETYPE     double
#define MPI_ETYPE MPI_DOUBLE 

main(int argc, char **argv)
{
  int      arg_error, MEMORY_PER_PROCESSOR, MT, SHARED, UNIQUE, DELETE; 
  int      ADAPT; 
  int      T_MAX, T_MIN, N_MAX, N_MIN, Ep_max;
  char     *str, *PATH, *PREFIX;
  double   T, Ta, Tn, Tu, Tp, sum_Tp, b_eff_io_Tn;
  double   t, t_pat, sum_t_pat, t_type, t_access, t_partition, t_total;
  double   t_sync, t_open, t_close, t_barrier, t_bcast;
  double   t_last_io, t_last_barr_bcast; 
  int      N,n, w_rank,n_rank, p, color, key, a,type,part,pos, r, r_max;
  int      i, l, L, L_m_part, U, Usum, l_etype, Ep, Ep_sum, En, En_sum;
  long     Lseg, Spat, Stype, Saccess;
  FILE     *fp, *fp_short, *fp_all[3], *fpx;
  MPI_Comm n_comm; 
  char     filename[50], full_filename[150]; 
  MPI_File fh; 
  char     conti; 
  int      gsizes, distribs, dargs;
  MPI_Datatype etype, filetype; 
  MPI_Offset   disp; 
  MPI_Status   sts; 
  ETYPE    *wr_buf, *rd_buf; 
  double   rep[PATTERNS];
  double   tt_pat[ACCESSES][PATTERNS];
  double   tt_type[ACCESSES][PAT_TYPES]; 
  double   bw_pat,  bw_type,  bw_access,  bw_partition;
  double           wbw_type, wbw_access, wbw_partition, b_eff_io; 
  int              wgh_type, wgh_access, wgh_partition, b_eff_io_n;
  double   bw_x[ACCESSES][PAT_TYPES][POSITIONS+1], wbw_x[ACCESSES];
  double   b_eff_io_ACCESSMETHOD_PATTERNTYPE [ACCESSES] [PAT_TYPES];
  double   b_eff_io_ACCESSMETHOD [ACCESSES], b_eff_io_PARTITION; 
  int      b_eff_io_En; 
  struct utsname  uname_info; 
  time_t   t_now; 
  int      crit1, crit2, crit3, crit4, crit_all, crit_all_n; 

  int l_pos[POSITIONS] = { 0, 1 kB, 1 kB +8, 32 kB, 32 kB +8,
                           1 MB, 1 MB +8, M_PART }; 

  char *access_string[ACCESSES] = { "write", "rewrite", "read" }; 
  char *pattype_string[PAT_TYPES]={ "scatter", "shared", "separated",
                                    "segmented", "seg_coll" }; 

  struct {int    type;  /* 0 = strided & individual pointer
                           1 = strided & shared pointer
                           2 = individual files, non-collective
                           3 = segmented files, non-collective
                           4 = segmented files, collective */
          int       l;
          int       L;
          int       U;
          int    part;  /* 0 = first,  1 = middle,  2 = last */
          int     pos;  /* 0 = unused, 1..7 = output line    */
         }             
      pattern_db[PATTERNS] =
       { /* type,        l,           L, U, part, pos */
         {     0,     1 MB,        1 MB, 0,    0,   0  },  
         {     0,   M_PART,        SAME, 4,    1,   7  },  
         {     0,     1 MB,        2 MB, 4,    1,   8  }, /*pos=8 - exception */ 
         {     0,     1 MB,        1 MB, 4,    1,   5  },  
         {     0,    32 kB,        1 MB, 2,    1,   3  },  
         {     0,     1 kB,        1 MB, 2,    1,   1  },  
         {     0, 32 kB +8, 1 MB + 256 , 2,    1,   4  },  
         {     0,  1 kB +8, 1 MB + 8 kB, 2,    1,   2  },  
         {     0,  1 MB +8, 1 MB +   8 , 2,    2,   6  },  
  
         {     1,     1 MB,        SAME, 0,    0,   0  },  
         {     1,   M_PART,        SAME, 4,    1,   7  },  
         {     1,     1 MB,        SAME, 2,    1,   5  },  
         {     1,    32 kB,        SAME, 1,    1,   3  },  
         {     1,     1 kB,        SAME, 1,    1,   1  },  
         {     1, 32 kB +8,        SAME, 1,    1,   4  },  
         {     1,  1 kB +8,        SAME, 1,    1,   2  },  
         {     1,  1 MB +8,        SAME, 2,    2,   6  },  
 
         {     2,     1 MB,        SAME, 0,    0,   0  },  
         {     2,   M_PART,        SAME, 2,    1,   7  },  
         {     2,     1 MB,        SAME, 2,    1,   5  },  
         {     2,    32 kB,        SAME, 1,    1,   3  },  
         {     2,     1 kB,        SAME, 1,    1,   1  },  
         {     2, 32 kB +8,        SAME, 1,    1,   4  },  
         {     2,  1 kB +8,        SAME, 1,    1,   2  },  
         {     2,  1 MB +8,        SAME, 2,    2,   6  },  
 
         {     3,     1 MB,        SAME, 0,    0,   0  },  
         {     3,   M_PART,        SAME, 2,    1,   7  },  
         {     3,     1 MB,        SAME, 2,    1,   5  },  
         {     3,    32 kB,        SAME, 1,    1,   3  },  
         {     3,     1 kB,        SAME, 1,    1,   1  },  
         {     3, 32 kB +8,        SAME, 1,    1,   4  },  
         {     3,  1 kB +8,        SAME, 1,    1,   2  },  
         {     3,  1 MB +8,        SAME, 2,    1,   6  },  
         {     3,        0,        SAME, 0,    2,   0  }, /*(1)*/
 
         {     4,     1 MB,        SAME, 0,    0,   0  },  
         {     4,   M_PART,        SAME, 2,    1,   7  },  
         {     4,     1 MB,        SAME, 2,    1,   5  },  
         {     4,    32 kB,        SAME, 1,    1,   3  },  
         {     4,     1 kB,        SAME, 1,    1,   1  },  
         {     4, 32 kB +8,        SAME, 1,    1,   4  },  
         {     4,  1 kB +8,        SAME, 1,    1,   2  },  
         {     4,  1 MB +8,        SAME, 2,    1,   6  },  
         {     4,        0,        SAME, 0,    2,   0  }  /*(1)*/
       };

  /* (1) part==2 also implies, that r_max = 1 is choosen */
 
  MPI_Init( &argc, &argv);
  MPI_Barrier(MPI_COMM_WORLD); 
  t_total = MPI_Wtime();
  MPI_Comm_size(MPI_COMM_WORLD, &N);
  MPI_Comm_rank(MPI_COMM_WORLD, &w_rank);
  /* defaults: */ 
    MEMORY_PER_PROCESSOR = -1; 
    MT = -1; 
#   ifdef WITHOUT_SHARED
      SHARED = 0; 
#   else
      SHARED = 1; 
#   endif 
    UNIQUE = 1; 
    DELETE = 1; 
    N_MAX  = N;
    N_MIN  = -1;  /* computed after the argument reading loop */ 
    T_MAX  = T_CRIT;
    T_MIN  = -1;  /* computed after the argument reading loop */ 
    PATH   = ".";  
    Ep_max = 1; 
    PREFIX = "b_eff_io"; 
    ADAPT  = 0; 
 
  arg_error = 0; 
  for (i=1; i < argc; i++) {
    str = argv[i];
    if (strcmp(str, "-MB") == 0) {
      i++;
      if (i < argc) {
         MEMORY_PER_PROCESSOR = atoi ( argv[i] ); 
         if (MEMORY_PER_PROCESSOR <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -MB\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -MB\n"); 
      } 
    }
    else if (strcmp(str, "-MT") == 0) {
      i++;
      if (i < argc) {
         MT = atoi ( argv[i] ); 
         if (MT <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -MT\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -MT\n"); 
      } 
    }
    else if (strcmp(str, "-noshared") == 0) {
      SHARED = 0; 
    }
    else if (strcmp(str, "-nounique") == 0) {
      UNIQUE = 0; 
    }
    else if (strcmp(str, "-keep") == 0) {
      DELETE = 0; 
    }
    else if (strcmp(str, "-d") == 0) {
      DELETE = 1; 
    }
    else if (strcmp(str, "-adapt") == 0) {
      ADAPT = 1; /* NOT ALLOWED FOR BENCHMARKING */
    }
    else if (strcmp(str, "-N") == 0) {
      i++;
      if (i < argc) {
         N_MAX = atoi ( argv[i] ); 
         if (N_MAX <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -N\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -N\n"); 
      } 
    }
    else if (strcmp(str, "-n") == 0) {
      i++;
      if (i < argc) {
         N_MIN = atoi ( argv[i] ); 
         if (N_MIN <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -n\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -n\n"); 
      } 
    }
    else if (strcmp(str, "-T") == 0) {
      i++;
      if (i < argc) {
         T_MAX = atoi ( argv[i] ); 
         if (T_MAX <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -T\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -T\n"); 
      } 
    }
    else if (strcmp(str, "-t") == 0) {
      i++;
      if (i < argc) {
         T_MIN = atoi ( argv[i] ); 
         if (T_MIN <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -t\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -t\n"); 
      } 
    }
    else if (strcmp(str, "-p") == 0) {
      i++;
      if (i < argc) {
         PATH = argv[i]; 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -p\n"); 
      } 
    }
    else if (strcmp(str, "-e") == 0) {
      i++;
      if (i < argc) {
         Ep_max = atoi ( argv[i] ); 
         if (Ep_max <= 0) { 
            arg_error++; 
            if (w_rank==0) printf( "Invalid argument for -e\n"); 
         } 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -e\n"); 
      } 
    }
    else if (strcmp(str, "-f") == 0) {
      i++;
      if (i < argc) {
         PREFIX = argv[i]; 
      } else { 
         arg_error++; 
         if (w_rank==0) printf( "Missing argument for -f\n"); 
      } 
    }
    else {
      arg_error++; 
      if (w_rank==0) printf( "Invalid option %s\n", str);
    } 
     
  } /* endfor (i=1; i < argc; i++) */  
  if (N_MIN == -1) N_MIN = N_MAX; 
  if (T_MIN == -1) T_MIN = T_MAX; 
  if (MEMORY_PER_PROCESSOR < 0) {
    arg_error++; 
    if (w_rank==0) 
       printf( "Missing mandatory argument -MB <memory per process> [in MBytes]\n");
  } 
  if (MT < 0) {
    arg_error++; 
    if (w_rank==0) 
       printf( "Missing mandatory argument -MT <total memory> [in MBytes]\n");
  } 
  if (arg_error > 0) {
    if (w_rank==0) printf( "Aborted due to argument errors.\n\n"); 
    if (w_rank==0) printf(HELP); 
    MPI_Finalize();
    return(1);
  }
   
  L_m_part = max(MEMORY_PER_PROCESSOR * 1024*(1024/128), 2 MB);
  if (w_rank == 0)
  { 
    sprintf(filename,"%s.prot",PREFIX); fp = fopen(filename,"w");
    fp_all[0] = fp;
    sprintf(filename,"%s.sum",PREFIX); fp_short = fopen(filename,"w");
    fp_all[1] = fp_short; 
    fp_all[2] = stdout;
    uname(&uname_info);
    t_now = time((time_t *) 0);
    for (i=0; i<2; i++)
    { fpx = fp_all[i]; 
      fprintf(fpx,"b_eff_io.c, Revision %s from %s\n\n", Revision, Date);
      fprintf(fpx,"MEMORY_PER_PROCESSOR = %4d MBytes  [1MBytes = 1024*1024 bytes, 1MB = 1e6 bytes]\n",
              MEMORY_PER_PROCESSOR);
      fprintf(fpx,"Maximum chunk size   = %8.3f MBytes\n\n",
              L_m_part/(1024.*1024.));
      fprintf(fpx,"N=%1d, n=%1d, T=%1d, t=%1d, MT=%1d MBytes",
                 N_MAX, N_MIN, T_MAX, T_MIN, MT);
#     ifdef WITHOUT_SHARED
        fprintf(fpx,", -D WITHOUT_SHARED");
#     endif 
      if (!SHARED) fprintf(fpx,", -noshared");
      if (!UNIQUE) fprintf(fpx,", -nounique");
      if (!DELETE) fprintf(fpx,", -keep");
      if (ADAPT)   fprintf(fpx,", -adapt (NOT ALLOWED FOR BENCHMARKING)");
      fprintf(fpx,"\n");
      fprintf(fpx,"PATH=%s, PREFIX=%s\n", PATH, PREFIX);
      fprintf(fpx,"       system name : %s\n", uname_info.sysname);
      fprintf(fpx,"       hostname    : %s\n", uname_info.nodename);
      fprintf(fpx,"       OS release  : %s\n", uname_info.release);
      fprintf(fpx,"       OS version  : %s\n", uname_info.version);
      fprintf(fpx,"       machine     : %s\n\n", uname_info.machine);
      fprintf(fpx,"Date of measurement: %s\n\n", ctime(&t_now));
    } 
  }

  etype = MPI_ETYPE;
  l_etype = sizeof(ETYPE); 

  if ( !( wr_buf = (ETYPE *)malloc((size_t)L_m_part) ) ||
       !( rd_buf = (ETYPE *)malloc((size_t)L_m_part) ) )
     { printf("[%d] ERROR - not enough space for buffers - "
              "is MEMORY_PER_PROCESSOR = %d MByte correct?\n",
              w_rank, MEMORY_PER_PROCESSOR);
       MPI_Abort(MPI_COMM_WORLD,1);
     }
  for (i=0; i<L_m_part/l_etype; i++) 
    { wr_buf[i] = -2*i; rd_buf[i] = -2*i-1; } 
 
  Usum = 0; 
  for (p=0; p<PATTERNS; p++)
  { if (pattern_db[p].l == M_PART) pattern_db[p].l = L_m_part;
    if (pattern_db[p].L == SAME)   pattern_db[p].L = pattern_db[p].l;
    pattern_db[p].L = pattern_db[p].L / l_etype; 
    pattern_db[p].l = pattern_db[p].l / l_etype; 
    Usum += pattern_db[p].U; 
  } 
  l_pos[7] = L_m_part; 

  T = T_MAX;

  b_eff_io = 0; 
  for (n = N_MAX; n >= max(1, N_MIN); n = n/2)
  {
    En = 0; 
    MPI_Barrier(MPI_COMM_WORLD); 
    t_partition = MPI_Wtime(); 
    Tn = max(min(T_MAX,T_MIN), T * n/N);
    Ta = Tn / ACCESSES;
    Tu = Ta / Usum; 
    color = (w_rank < n ? 1 : 2);  key = 0; 
    MPI_Comm_split(MPI_COMM_WORLD,color,key,&n_comm); 
    MPI_Comm_rank(n_comm, &n_rank);
    MPI_Barrier(MPI_COMM_WORLD); 
    if (color==1) 
    {
      if (w_rank != n_rank) printf("[%d] ERROR - ranks %d!=!d\n",
                                   w_rank, w_rank, n_rank); 
      if (w_rank == 0) {
        fprintf(fp,"num.  ac-  pat-  pat- scheduled     chunk      chunk filename          repeat   transferred   meas- =sum of time of meas.calls|last  last | measured       \n");
        fprintf(fp,"of   cess  tern  tern   time         size       size                   factor    MB of this   ured  |  I/O |barr- |bcast|file-| I/O  barr.| bandwidth      \n");
        fprintf(fp,"PEs        type        [sec]      on disk  in memory                                pattern   time  |      | ier  |     | sync|call  +bcst| of this pattern\n");
        fprintf(fp,"----- --- ------ ---- -------- ---------- ---------- ---------------- ------- ------------- -------  ------ ------ ----- ----- ----- ----- ----------------\n");
      } 
      bw_partition = 0; 
      wbw_partition = 0; wgh_partition = 0;
      for (a=0; a<ACCESSES; a++)
      {
        t_access = MPI_Wtime(); 
        Saccess = 0; 
        b_eff_io_ACCESSMETHOD[a] = 0; 
        wbw_access = 0; wgh_access = 0;
        sum_Tp = 0;     sum_t_pat = 0; 
        for (p=0; p<PATTERNS; p++)
        { Ep = 0;
          type = pattern_db[p].type; 
          part = pattern_db[p].part; 
          pos  = pattern_db[p].pos; 
          if (type==3 && part==0)
          { /* Initializing the repetition counts and Lseg for type=3,4 */
            int p0, p1, p2, p3; 
            long L3; 
            L3 = 0; 
            rep[TYPE_3_START] = 1;
            rep[TYPE_4_START] = 1;
            L3 += pattern_db[TYPE_3_START].L; 
            for (p3=1; p3<TYPE_3_PTRNS-1; p3++)
            { p0 = (p3 <= 1 ? p3 : p3+1);    p1 = p3;   p2 = p3;
              rep[TYPE_3_START+p3] =
                min(rep[TYPE_0_START+p0]
                     *(pattern_db[TYPE_0_START+p0].L 
                       / pattern_db[TYPE_0_START+p0].l),
                 min(10*rep[TYPE_1_START+p1],
                  rep[TYPE_2_START+p2] ));    
              L3 += rep[TYPE_3_START+p3] * pattern_db[TYPE_3_START+p3].L;
              rep[TYPE_4_START+p3] = rep[TYPE_3_START+p3]; 
            } 
            Lseg = ( (L3-1) / (1 MB / l_etype) + 1) * (1 MB / l_etype);
            pattern_db[TYPE_3_START+TYPE_3_PTRNS-1].l = (int)(Lseg-L3);
            pattern_db[TYPE_3_START+TYPE_3_PTRNS-1].L = (int)(Lseg-L3);
            rep[TYPE_3_START+TYPE_3_PTRNS-1] = 1; 
            pattern_db[TYPE_4_START+TYPE_4_PTRNS-1].l = (int)(Lseg-L3);
            pattern_db[TYPE_4_START+TYPE_4_PTRNS-1].L = (int)(Lseg-L3);
            rep[TYPE_4_START+TYPE_4_PTRNS-1] = 1; 
          } 
          l  = pattern_db[p].l; 
          L  = pattern_db[p].L; 
          U  = pattern_db[p].U; 
          Tp = Tu * U;
          if (ADAPT && a==0 && type<=2 && p>=1 
                    && sum_t_pat>sum_Tp && U!=0 && sum_Tp>0)
               /* reduction if t_sync results in sum_t_pat > sum_Tp */ 
               Tp = Tp / ( 0.5 * (1 + sum_t_pat/sum_Tp) );
          if (a>0 || type>=3) r_max = rep[p];
          else r_max = 0; 
          Spat = 0; 
          if (part == 0)
          {
            Stype = 0; 
            wbw_type = 0; wgh_type = 0;
            if (type != 2) 
              sprintf(filename,"b_eff_io_data_%03d_%03d_%1d",N,n,type); 
            else 
              sprintf(filename,"b_eff_io_data_%03d_%03d_%1d.%03d",N,n,type,n_rank); 
            sprintf(full_filename,"%s/%s",PATH,filename); 
            if (a==0)
            {  /* remove existing old file */
               MPI_File_open((type != 2 ? n_comm : MPI_COMM_SELF),
                             full_filename,
                             MPI_MODE_CREATE|MPI_MODE_RDWR|
                               MPI_MODE_DELETE_ON_CLOSE,
                             MPI_INFO_NULL,&fh); 
               MPI_File_close(&fh); 
            } 
            MPI_Barrier(n_comm); 
            t_type = 0;
            t_open = MPI_Wtime(); 
            MPI_File_open((type != 2 ? n_comm : MPI_COMM_SELF),
                          full_filename,
                          (a==0 ? MPI_MODE_CREATE | MPI_MODE_RDWR
                                :(a==1 ? MPI_MODE_RDWR
                                       :/*=2*/ MPI_MODE_RDONLY
                                             | (DELETE ? MPI_MODE_DELETE_ON_CLOSE : 0)
                                  ))
                           | (UNIQUE ? MPI_MODE_UNIQUE_OPEN : 0)
                          ,MPI_INFO_NULL, &fh); 
            MPI_Barrier(n_comm); 
            t_open = MPI_Wtime() - t_open;
            t_type += t_open; 
          } /*endif (part == 0)*/
 
          MPI_Barrier(n_comm); 
          t_pat = -MPI_Wtime(); 
          t_barrier = 0;  t_bcast = 0;  t_sync = 0;

          if (type == 0 
              || ((type == 1) && (!SHARED)) 
             ) 
          { 
            gsizes   = l * n; 
            distribs = MPI_DISTRIBUTE_BLOCK;
            dargs    = MPI_DISTRIBUTE_DFLT_DARG;
            MPI_Type_create_darray(n, n_rank, 1, &gsizes, &distribs,
                                   &dargs, &n, MPI_ORDER_C,etype,
                                   &filetype);
            MPI_Type_commit(&filetype);
            if (part == 0) disp = 0; 
            MPI_File_set_view(fh, disp, etype, filetype, "native", 
                              MPI_INFO_NULL); 
            MPI_Type_free(&filetype); 
          } 
          else if (part == 0) 
          { 
            if (type<3) disp = 0;
            else        disp = ((MPI_Offset)n_rank) * ((MPI_Offset)Lseg)
                               * ((MPI_Offset)l_etype);
            MPI_File_set_view(fh, disp, etype, etype,
                                             "native", MPI_INFO_NULL); 
          } 
 
          r = 0;
          do { 
            t_last_io = -MPI_Wtime(); 
            if (a<2) for (i=0; i<L; i+=l)
                     { wr_buf[i]     = 1e7*r + 1000*i + 10*p + a+1;
                       wr_buf[i+l-1] = 1e7*r + 1000*i + 10*p + a+6;
                     } 
            switch (type)
            {
             case 0: /* strided, individual pointer, collective */ 
              if(a<2) MPI_File_write_all(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read_all (fh,rd_buf,L,etype,&sts);
             break; 
             case 1: /* strided, shared pointer, collective */ 
#ifdef WITHOUT_SHARED
                     /* substituted individual pointer */
              if(a<2) MPI_File_write_all(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read_all (fh,rd_buf,L,etype,&sts);
#else 
              if(a<2) MPI_File_write_ordered(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read_ordered (fh,rd_buf,L,etype,&sts);
#endif 
             break; 
             case 2: /* individual files, non-collective */ 
              if(a<2) MPI_File_write(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read (fh,rd_buf,L,etype,&sts);
             break; 
             case 3: /* segmented, individual pointer, collective */ 
              if(a<2) MPI_File_write(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read (fh,rd_buf,L,etype,&sts);
             break; 
             case 4: /* segmented, individual pointer, collective */ 
              if(a<2) MPI_File_write_all(fh,wr_buf,L,etype,&sts);
              else    MPI_File_read_all (fh,rd_buf,L,etype,&sts);
             break; 
            } 
            /* if (a==2 && p==7 && w_rank<2) rd_buf[0]=-99999; *TEST*/
            if (a==2) for (i=0; i<L; i+=l) 
              if (   (rd_buf[i]     != 1e7*r + 1000*i + 10*p + 1+1)
                   ||(rd_buf[i+l-1] != 1e7*r + 1000*i + 10*p + 1+6) )
                { Ep++;  En++;
                  /* TEST: the following conditions can be modified
                           without real impact to the total result: */
                  if (Ep<=Ep_max)
                    printf("[%d] ERROR in reading: p=%d, r=%d, i=%d, buf=(%1.0f,%1.0f) != (%1.0f,%1.0f)\n",
                         w_rank,p,r,i,rd_buf[i],rd_buf[i+l-1],
                         1e7*r + 1000*i + 10*p + 1+1, 
                         1e7*r + 1000*i + 10*p + 1+6);
                  if (Ep<=Ep_max && w_rank==0) 
                    fprintf(fp,"[%d] ERROR in reading: p=%d, r=%d, i=%d, buf=(%1.0f,%1.0f) != (%1.0f,%1.0f)\n",
                         w_rank,p,r,i,rd_buf[i],rd_buf[i+l-1],
                         1e7*r + 1000*i + 10*p + 1+1, 
                         1e7*r + 1000*i + 10*p + 1+6);
                } 
            r++; 
            t = MPI_Wtime();
            t_last_io += t; 
            t_last_barr_bcast = -t; 
            if (a==0 && type<=2)
            {
              t_barrier -= t;
              MPI_Barrier(n_comm); 
              t = MPI_Wtime(); 
              t_barrier += t; 
              t_bcast -= t; 
              conti = (t + t_pat) * (r+1) / r <= Tp;
              MPI_Bcast(&conti, 1, MPI_CHAR, 0, n_comm); 
              t = MPI_Wtime(); 
              t_bcast += t;
            } 
            t_last_barr_bcast += t; 
          } while((a==0 && type<=2 ? conti : r<r_max)); 
          Spat = l_etype * L * r;
          Stype += Spat;  Saccess += Spat; 
          if(a<2) 
          { t_barrier -= MPI_Wtime(); 
            MPI_Barrier(n_comm); 
            t = MPI_Wtime(); 
            t_barrier += t; 
            t_sync -= t;
            MPI_File_sync(fh); 
            t_sync += MPI_Wtime(); 
          } 
          t_barrier -= MPI_Wtime(); 
          MPI_Barrier(n_comm); 
          t = MPI_Wtime(); 
          t_barrier += t; 
          t_pat += t;
          t_type += t_pat;
          rep[p] = r;
          tt_pat[a][p] = t_pat;
          bw_pat = Spat / t_pat * n; 
          wbw_type += bw_pat * U;       wgh_type += U; 
          wbw_access += bw_pat * U;     wgh_access += U; 
          wbw_partition += bw_pat * U;  wgh_partition += U; 
          if (type == 0 
              || ((type == 1) && (!SHARED)) 
              ) disp = disp + ((MPI_Offset)n) * ((MPI_Offset)Spat);
          sum_Tp += Tu * U;  sum_t_pat += t_pat; 
          if (part == 2)
          {
            t_close = MPI_Wtime();
            MPI_File_close(&fh); 
            MPI_Barrier(n_comm); 
            t = MPI_Wtime(); 
            t_close = t - t_close;
            t_type  += t_close;
            tt_type[a][type] = t_type;
            bw_type = Stype / t_type * n; 
          } 
          MPI_Reduce(&Ep, &Ep_sum, 1, MPI_INT, MPI_SUM, 0, n_comm);  
          if (w_rank == 0) 
          { 
            fprintf(fp,"n=%-3d a=%d type=%d p=%2d Tp=%5.2f l=%8ld L=%8ld %-16s r=%5d S=%8.3f MB t=%5.2f =%6.3f+%6.3f+%5.3f+%5.3f %5.3f %5.3f bw=%8.3f MB/s",
                         n,a,type,p,Tp,l*l_etype,L*l_etype,filename+14,
                         r,1e-6*Spat*n,t_pat,
                         t_pat - t_barrier - t_bcast - t_sync,
                         t_barrier,t_bcast,t_sync,
                         t_last_io, t_last_barr_bcast,  1e-6*bw_pat);
            if (Ep_sum > 0) fprintf(fp," %d ERRORS",Ep_sum); 
            fprintf(fp,"\n"); fflush(fp);
            bw_x[a][type][pos] = ( Ep_sum > 0 ? -999.999e6 : bw_pat ); 
            if (part == 2) 
            { fprintf(fp, " total pattern type:  S=%7.3f MB  t=%4.2f t_op=%4.2f t_cl=%4.2f wbw=%8.3f MB/s b_eff_io_%s_%s=%8.3f MB/s",
                          1e-6*Stype*n,t_type,t_open,t_close, 
                          1e-6*wbw_type/wgh_type, 
                          access_string[a],pattype_string[type],
                          1e-6*bw_type );
              if (type>=3) fprintf(fp, " Lseg=%ld MBytes",l_etype*Lseg/(1 MB)); 
              fprintf(fp, "\n"); fflush(fp);
              b_eff_io_ACCESSMETHOD_PATTERNTYPE[a][type] = bw_type;
              b_eff_io_ACCESSMETHOD[a] += (type==0 ? 2 : 1) * bw_type;
                               /* double weighted scatter-type 0 */
            } 
          } 
        } /*endfor p*/
        MPI_Barrier(n_comm); 
        t_access = MPI_Wtime() - t_access; 
        bw_access = Saccess / t_access * n; 
        wbw_x[a] = wbw_access/wgh_access;
        b_eff_io_ACCESSMETHOD[a] = b_eff_io_ACCESSMETHOD[a] 
                                   / (PAT_TYPES + 1); 
             /* double weighted scatter-type 0: + 1 */
        if (w_rank == 0) {
          fprintf(fp, " total access method: S=%7.3f MB  t=%4.2f bw=%8.3f MB/s, wbw=%8.3f MB/s, b_eff_io_%s=%8.3f MB/s\n",
                      1e-6*Saccess*n,t_access,bw_access*1e-6, 
                      1e-6*wbw_access/wgh_access,
                      access_string[a], 1e-6*b_eff_io_ACCESSMETHOD[a] );
          fflush(fp); 
        } 
      } /*endfor a*/
      MPI_Barrier(n_comm); 
      t_partition = MPI_Wtime() - t_partition; 
      bw_partition = Saccess / t_partition * ACCESSES * n; 
      b_eff_io_PARTITION =  0.25 * b_eff_io_ACCESSMETHOD[0 /*write*/  ] 
                          + 0.25 * b_eff_io_ACCESSMETHOD[1 /*rewrite*/] 
                          + 0.50 * b_eff_io_ACCESSMETHOD[2 /*read*/   ]; 
      MPI_Reduce(&En, &En_sum, 1, MPI_INT, MPI_SUM, 0, n_comm);  
      if (w_rank == 0) 
      { fprintf(fp, " total partition:     S=%7.3f MB  t=%4.2f (Tn=%1.0fsec) bw=%8.3f MB/s, wbw=%8.3f MB/s, b_eff_io=%8.3f MB/s",
                    1e-6*ACCESSES*Saccess*n,t_partition,Tn,
                    1e-6*bw_partition, 1e-6*wbw_partition/wgh_partition,
                    1e-6*b_eff_io_PARTITION );
        if (En_sum > 0) fprintf(fp," ----- with %d ERRORS --!--!--!--!--",En_sum); 
        fprintf(fp,"\n\n"); fflush(fp);
        fprintf(fp_short, "\nSummary of file I/O bandwidth accumulated on %3d processes with %4d MByte/PE\n", n, MEMORY_PER_PROCESSOR); 
        fprintf(fp_short,   "-----------------------------------------------------------------------------\n\n"); 
        fprintf(fp_short, " number pos chunk-   access   type=0   type=1   type=2   type=3   type=4\n");
        fprintf(fp_short, " of PEs     size (l) methode scatter  shared  separate segmened seg-coll\n");
        fprintf(fp_short, "            [bytes]  methode  [MB/s]   [MB/s]   [MB/s]   [MB/s]   [MB/s]\n");
        fprintf(fp_short, " -----------------------------------------------------------------------\n\n"); 
        for (a=0; a<ACCESSES; a++)
        { 
          for (pos=1; pos<POSITIONS; pos++)
          {
            fprintf(fp_short, "%4d PEs %1d %8d %-7s",
                              n,pos,l_pos[pos],access_string[a]); 
            for (type=0; type<PAT_TYPES; type++)
              fprintf(fp_short, " %8.3f",bw_x[a][type][pos]*1e-6); 
            fprintf(fp_short, "\n"); 
          }   
          fprintf(fp_short, "%4d PEs      total-%-7s",n,access_string[a]);
          for (type=0; type<PAT_TYPES; type++)
            fprintf(fp_short, " %8.3f",
                    b_eff_io_ACCESSMETHOD_PATTERNTYPE[a][type]*1e-6); 
          fprintf(fp_short, "\n"); 
          fprintf(fp_short, "\n"); 
        }   
        fprintf(fp_short, "This table shows all results, except pattern 2 (scatter, l=1MBytes, L=2MBytes): \n");
        fprintf(fp_short, "  bw_pat2= %8.3f MB/s write, %8.3f MB/s rewrite, %8.3f MB/s read\n\n",
                bw_x[0][0][8]*1e-6, bw_x[1][0][8]*1e-6, bw_x[2][0][8]*1e-6);
        fprintf(fp_short, "(For gnuplot:)\n");
        fprintf(fp_short, "  set xtics ( '1k' 1, '+8' 2, '32k' 4, '+8' 5, '1M' 7, '+8' 8, '%1dM' 10)\n",L_m_part/1024/1024);
        fprintf(fp_short, "  set title '%s %s %s %s %s' -4\n",
            uname_info.sysname, uname_info.nodename, uname_info.release,
            uname_info.version, uname_info.machine);
        fprintf(fp_short, "  set label 1 'b_eff_io'  at 10,50000 right\n");
        fprintf(fp_short, "  set label 2 'rel. %s'  at 10,25000 right\n",Revision);
        fprintf(fp_short, "  set label 3 'T=%3.1fmin' at 10,12500 right\n",Tn/60.);
        fprintf(fp_short, "  set label 4 'n=%1d'     at 10,6250  right\n",n);
        if (!SHARED) { 
          fprintf(fp_short,"  set label 5 'workaround for type 1:'  at 10,0.50 right\n");
          fprintf(fp_short,"  set label 6 'individual file pointer' at 10,0.25 right\n");
        } 
        fprintf(fp_short, "\n");
        for (a=0; a<ACCESSES; a++)
        {
           fprintf(fp_short,"weighted average bandwidth for %-7s : %8.3f MB/s on %1d processes\n",
                             access_string[a], 
                             b_eff_io_ACCESSMETHOD[a]*1e-6, n);
        }
        fprintf(fp_short,"(type=0 is weighted double)\n");
        fprintf(fp_short, "\nTotal amount of data written/read with each access method: %8.3f MBytes\n", 1.0 * Saccess*n / (1 MB));
        fprintf(fp_short, "  = %4.1f percent of the total memory (%1d MBytes)\n", 100.*Saccess*n/(MT * 1.0 MB), MT);
        fprintf(fp_short, "\nb_eff_io of these measurements = %8.3f MB/s on %1d processes with %3d MByte/PE and scheduled time=%3.1f min",
                          1e-6*b_eff_io_PARTITION, n, 
                          MEMORY_PER_PROCESSOR, Tn/60.); 
        if (En_sum > 0) fprintf(fp_short,", with %d ERRORS --!--!--!--",En_sum);
        fprintf(fp_short, "\n\n"); 
 
        crit1 = (Tn >= T_CRIT); 
        crit2 = (Saccess*n >= T_PERCENTAGE * (MT * 1.0 MB));
        crit3 = (SHARED == 1);
        crit4 = (En_sum == 0); 
        crit_all = crit1 && crit2 && crit3 && crit4; 
        fprintf(fp_short, "%s for comparison of different systems\n",
                (crit_all ? "VALID" : "NOT VALID") );
        if ( ! crit_all ) { 
          fprintf(fp_short, "  criterion 1: scheduled time %3.1f min >= %2d min -- %s\n", 
                  Tn/60., T_CRIT/60, (crit1 ? "reached" : "NOT reached"));
          fprintf(fp_short, "  criterion 2: transferred data / total memory = %5.1f %% >= %2.0f %% -- %s\n",
                  100. * Saccess*n / (MT * 1.0 MB), 100.*T_PERCENTAGE,
                  (crit2 ? "reached" : "NOT reached") );
          fprintf(fp_short, "  criterion 3: shared file pointers must be used for pattern type 1 -- %s\n",
                  (crit3 ? "reached" : "NOT reached") );
          fprintf(fp_short, "  criterion 4: error count (%1d) == 0 -- %s\n",
                  En_sum, (crit4 ? "reached" : "NOT reached") ); 
        } 
 
        if (b_eff_io_PARTITION > b_eff_io)
        { b_eff_io   = b_eff_io_PARTITION;
          b_eff_io_n = n;  b_eff_io_Tn = Tn;  b_eff_io_En = En_sum;
          crit_all_n = crit_all; 
        } 
      } 
    } /*endif (color==1)*/ 
    MPI_Barrier(MPI_COMM_WORLD); 
  } /*endfor n*/
  MPI_Barrier(MPI_COMM_WORLD); 
  t_total = MPI_Wtime() - t_total;
  if (w_rank == 0) 
  { 
    fprintf(fp, " total execution time is %7.2f sec\n",t_total); 
    fprintf(fp, "\n b_eff_io := maximum over all bw_partition = %8.3f MB/s\n",b_eff_io*1e-6 );
    fprintf(fp_short, "\nMaximum over all number of PEs\n"); 
    fprintf(fp_short,   "------------------------------\n\n"); 
 
    for (i=0; i<3; i++)
    { fpx = fp_all[i]; 
      fprintf(fpx, "\n b_eff_io = %8.3f MB/s on %1d processes"
                   " with %3d MByte/PE, scheduled time=%3.1f Min, on %s %s %s %s %s",
        b_eff_io*1e-6, b_eff_io_n, MEMORY_PER_PROCESSOR, b_eff_io_Tn/60,
        uname_info.sysname, uname_info.nodename, uname_info.release,
        uname_info.version, uname_info.machine);
      if (!crit_all_n) fprintf(fpx,", NOT VALID (see above)"); 
      if (b_eff_io_En > 0) fprintf(fpx,", with %d ERRORS --!--!--!--",b_eff_io_En);
      fprintf(fpx,"\n"); 
    } 
    fclose(fp); 
    fclose(fp_short); 
  } 
 
  MPI_Finalize();
  return(0);
} 
